//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		A BSlider-like control with two knobs.

#include "IntervalSlider.h"

#include <cstring>

#include <Bitmap.h>
#include <Font.h>
#include <GraphicsDefs.h>
#include <InterfaceDefs.h>
#include <Message.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <Slider.h>
#include <SupportDefs.h>
#include <View.h>

// Constant data.
const uint32 kDefaultHashMarkCount = 10;

const float kDefaultBarThickness = 3.0;
const float kHashMarkSize = 6.0;
const float kThumbWidth = 10.0;
const float kThumbHeight = 14.0;

const rgb_color kWhite = { 255, 255, 255, 255 }; 
const rgb_color kLightGray = { 235, 235, 235, 255 }; 
const rgb_color kMediumGray = { 192, 192, 192, 255 }; 
const rgb_color kDarkGray = { 144, 144, 144, 255 }; 
const rgb_color kBlack = { 0, 0, 0, 255};
const rgb_color kDefaultBarColor = kBlack;

static const char kFieldLowValue[]  = "jjk:low_value";
static const char kFieldHighValue[] = "jjk:high_value";

IntervalSlider::IntervalSlider(
	BRect frame, const char *pName, const char *pLabel,
	BMessage *pMsg, int16 minValue, int16 maxValue,
	uint32 resizingMode, uint32 flags)
:	inherited(frame, pName, pLabel, pMsg, resizingMode,
				 flags|B_WILL_DRAW|B_FRAME_EVENTS|B_FULL_UPDATE_ON_RESIZE),
	mMinValue(minValue), mMaxValue(maxValue),
	mpModificationMessage(0),
	mTracking(NotTracking),
	mTrackingOffset(B_ORIGIN),
	mpMinLabel(0),
	mpMaxLabel(0),
	mHashMarks(B_HASH_MARKS_NONE),
	mHashMarkCount(kDefaultHashMarkCount),
	mBarThickness(kDefaultBarThickness),
	mLowBarColor(kDefaultBarColor),
	mMidBarColor(kDefaultBarColor),
	mHighBarColor(kDefaultBarColor),
	mpOffscreenBits(0),
	mpOffscreenView(0),
	mViewColor(kMediumGray),
	mFont(be_plain_font)
{
	if (minValue >= maxValue) {
		debugger("Minimum value must be smaller than maximum.");
	}

	// Finish initialization of members.
	mFont.GetHeight(&mFontHeight);
	
	// Ensure valid values.
	SetLowValue(minValue);
	SetHighValue(maxValue);
}

IntervalSlider::~IntervalSlider(void)
{
	delete mpModificationMessage;
	mpModificationMessage = 0;
	delete [] mpMinLabel;
	mpMinLabel = 0;
	delete [] mpMaxLabel;
	mpMaxLabel = 0;
	DestroyOffscreenView();
}

void
IntervalSlider::AttachedToWindow(void)
{
	inherited::AttachedToWindow();
	
	// The view color has been cached for offscreen drawing;
	// set "real" view color to transparent to avoid redraw flicker.
	inherited::SetViewColor(B_TRANSPARENT_COLOR);
	
	CreateOffscreenView();
}

void
IntervalSlider::DetachedFromWindow(void)
{
	DestroyOffscreenView();
	SetViewColor(mViewColor);
	inherited::DetachedFromWindow();
}

void
IntervalSlider::Draw(BRect updateRect)
{
	// Draw into offscreen buffer, then copy to screen.
	if (mpOffscreenBits->Lock()) {
		// Ensure consistent graphical state.
		mpOffscreenView->SetViewColor(mViewColor);
		mpOffscreenView->SetLowColor(mViewColor);
		mpOffscreenView->SetFont(&mFont);
		
		DrawSlider();
		mpOffscreenView->Sync();
		mpOffscreenBits->Unlock();
	}
	DrawBitmap(mpOffscreenBits, updateRect, updateRect);
}

status_t
IntervalSlider::Invoke(BMessage *pMsg)
{
	// Copy the given message, or the model message if none was given.
	BMessage copy(pMsg ? *pMsg : *Message());
	
	// Add low and high values to the message (not strictly necessary,
	// but convenient).
	status_t status = copy.AddInt16(kFieldLowValue, LowValue());
	if (status == B_OK) {
		status = copy.AddInt16(kFieldHighValue, HighValue());
	}
	
	// Pass the message to the base class.
	if (status == B_OK) {
		status = inherited::Invoke(&copy);
	}
	return status;
}

void
IntervalSlider::SetModificationMessage(BMessage *pMsg)
{
	delete mpModificationMessage;
	mpModificationMessage = pMsg;
}

BMessage *
IntervalSlider::ModificationMessage(void) const
{
	return mpModificationMessage;
}

void
IntervalSlider::FrameResized(float, float)
{
	// Resize the offscreen buffer.
	DestroyOffscreenView();
	CreateOffscreenView();
}

void
IntervalSlider::GetPreferredSize(float *pWidth, float *pHeight)
{
	// The slider can have any width, provided the thumbs will fit.
	const float minWidth = 2.0 * kThumbWidth + 4.0 * mHashMarkCount;
	const float width = Bounds().Width();
	*pWidth = (width < minWidth) ? minWidth : width;

	// The height depends on the current font.
	const float lineHeight = mFontHeight.ascent
								  + mFontHeight.descent
								  + mFontHeight.leading;
	
	*pHeight = lineHeight				// label and status text
				+ kHashMarkSize			// top hash marks
				+ mBarThickness + 3.0	// bar, adjusted for bevels
				+ kHashMarkSize			// bottom hash marks
				+ lineHeight;				// limit labels
}

void
IntervalSlider::MouseDown(BPoint where)
{
	// Allow no interaction while disabled.
	if (!IsEnabled()) {
		return;
	}
	
	const BRect lowFrame = LowThumbFrame();
	const BRect highFrame = HighThumbFrame();
	
	if (lowFrame.Contains(where)) {
		// Start dragging the low thumb.
		mTracking = LowTracking;
		mTrackingOffset = lowFrame.RightTop() - where;

	} else if (highFrame.Contains(where)) {
		// Start dragging the high thumb.
		mTracking = HighTracking;
		mTrackingOffset = highFrame.LeftTop() - where;
	}
	
	if (mTracking != NotTracking) {
		// Start tracking mouse movement.
		SetMouseEventMask(B_POINTER_EVENTS,
								B_LOCK_WINDOW_FOCUS|B_NO_POINTER_HISTORY);
	}
}

void
IntervalSlider::MouseMoved(BPoint where, uint32, const BMessage *)
{
	// Allow no interaction while disabled.
	if (!IsEnabled()) {
		return;
	}
	
	switch (mTracking) {
	
	case NotTracking:
		// do nothing
		break;
	
	case LowTracking:
		// Update the low value.
		SetLowValue(ValueForPoint(where + mTrackingOffset));
		break;
	
	case HighTracking:
		// Update the high value.
		SetHighValue(ValueForPoint(where + mTrackingOffset));
		break;
	}
	
	if (mTracking != NotTracking && mpModificationMessage) {
		// Send the modification message.
		InvokeNotify(mpModificationMessage, B_CONTROL_MODIFIED);
	}
}

void
IntervalSlider::MouseUp(BPoint)
{
	// Allow no interaction while disabled.
	if (!IsEnabled()) {
		return;
	}
	
	if (mTracking != NotTracking) {
		// Send the final message.
		mTracking = NotTracking;
		InvokeNotify(Message());
	}
}

void
IntervalSlider::GetLimits(int32 &minimum, int32 &maximum) const
{
	minimum = mMinValue;
	maximum = mMaxValue;
}

void
IntervalSlider::SetValue(int32 value)
{
	// Ensure that both sub-values are legal.
	// No attempt is made to correct wrong values.
	int16 lowValue  = (int16)(value & 0x0000ffff);
	int16 highValue = (int16)((value >> 16) & 0x0000ffff);

	if (lowValue < mMinValue || lowValue > mMaxValue
	    || highValue < mMinValue || highValue > mMaxValue
	    || lowValue > highValue) {
		debugger("Invalid argument to SetValue().");
	}
	
	inherited::SetValue(value);
}

void
IntervalSlider::SetLowValue(int16 value)
{
	// Ensure value is within bounds and low value stays below high value.
	value = ClampValue(value);
	if (value > HighValue()) {
		inherited::SetValue(((int32)value << 16) | ((int32)value & 0x0000ffff));
	} else {
		inherited::SetValue((Value() & 0xffff0000) | ((int32)value & 0x0000ffff));
	}
}

int16
IntervalSlider::LowValue(void) const
{
	return (int16)(Value() & 0x0000ffff);
}

void
IntervalSlider::SetHighValue(int16 value)
{
	// Ensure value is within bounds and low value stays below high value.
	value = ClampValue(value);
	if (value < LowValue()) {
		inherited::SetValue(((int32)value << 16) | ((int32)value & 0x0000ffff));
	} else {
		inherited::SetValue(((int32)value << 16) | (Value() & 0x0000ffff));
	}
}

int16
IntervalSlider::HighValue(void) const
{
	return (int16)((Value() >> 16) & 0x0000ffff);
}

void
IntervalSlider::SetLowPosition(float position)
{
	SetLowValue(PositionToValue(position));
}

float
IntervalSlider::LowPosition(void) const
{
	return ValueToPosition(LowValue());
}

void
IntervalSlider::SetHighPosition(float position)
{
	SetHighValue(PositionToValue(position));
}

float
IntervalSlider::HighPosition(void) const
{
	return ValueToPosition(HighValue());
}

void
IntervalSlider::SetLimitLabels(const char *pMinLabel, const char *pMaxLabel)
{
	bool changed = false;
	
	// Set new minimum label if changed.
	if (!pMinLabel != !mpMinLabel
		 || (pMinLabel && strcmp(pMinLabel, mpMinLabel))) {
		delete [] mpMinLabel;
		mpMinLabel = pMinLabel
					  ? strcpy(new char[strlen(pMinLabel)+1], pMinLabel)
					  : 0;
		changed = true;
	}
	
	// Set new maximum label if changed.
	if (!pMaxLabel != !mpMaxLabel
		 || (pMaxLabel && strcmp(pMaxLabel, mpMaxLabel))) {
		delete [] mpMaxLabel;
		mpMaxLabel = pMaxLabel
					  ? strcpy(new char[strlen(pMaxLabel)+1], pMaxLabel)
					  : 0;
		changed = true;
	}
	
	if (changed) {
		Invalidate();
	}
}

const char *
IntervalSlider::MinLimitLabel(void) const
{
	return mpMinLabel;
}

const char *
IntervalSlider::MaxLimitLabel(void) const
{
	return mpMaxLabel;
}

void
IntervalSlider::SetHashMarks(hash_mark_location where)
{
	if (where != mHashMarks) {
		mHashMarks = where;
		Invalidate();
	}
}

hash_mark_location
IntervalSlider::HashMarks() const
{
	return mHashMarks;
}

void
IntervalSlider::SetHashMarkCount(uint32 count)
{
	if (count != mHashMarkCount) {
		mHashMarkCount = count;
		Invalidate();
	}
}

uint32
IntervalSlider::HashMarkCount(void) const
{
	return mHashMarkCount;
}

void
IntervalSlider::SetBarThickness(float thickness)
{
	if (thickness < 1.0) {
		debugger("Invalid bar thickness.");
	}
	if (thickness != mBarThickness) {
		mBarThickness = thickness;
		Invalidate();
	}
}

float
IntervalSlider::BarThickness(void) const
{
	return mBarThickness;
}

void
IntervalSlider::SetBarColors(rgb_color low, rgb_color mid, rgb_color high)
{
	bool changed = false;

	// Set low color if changed.
	if (memcmp(&low, &mLowBarColor, sizeof(rgb_color))) {
		mLowBarColor = low;
		changed = true;
	}
	// Set midrange color if changed.
	if (memcmp(&mid, &mMidBarColor, sizeof(rgb_color))) {
		mMidBarColor = mid;
		changed = true;
	}
	// Set high color if changed.
	if (memcmp(&high, &mHighBarColor, sizeof(rgb_color))) {
		mHighBarColor = high;
		changed = true;
	}

	if (changed) {
		Invalidate();
	}
}

rgb_color
IntervalSlider::LowBarColor(void) const
{
	return mLowBarColor;
}

rgb_color
IntervalSlider::MidBarColor(void) const
{
	return mMidBarColor;
}

rgb_color
IntervalSlider::HighBarColor(void) const
{
	return mHighBarColor;
}

void
IntervalSlider::SetFont(const BFont *pFont, uint32 properties)
{
	inherited::SetFont(pFont, properties);
	
	// Update cached font.
	GetFont(&mFont);
	mFont.GetHeight(&mFontHeight);

	Invalidate();
}

void
IntervalSlider::SetViewColor(rgb_color viewColor)
{
	mViewColor = viewColor;
	Invalidate();
}

int16
IntervalSlider::ValueForPoint(BPoint where) const
{
	BRect barFrame = BarFrame();
	barFrame.left += 2;	// adjust for bevels
	--barFrame.right;

	// Compute relative position; may lie outside the range [0.0, 1.0].
	const float position =
		(where.x - barFrame.left) / (barFrame.right - barFrame.left);
	// Ensure returned value lies within the legal range.	
	return ClampValue(PositionToValue(position));
}

BView &
IntervalSlider::OffscreenView(void) const
{
	return *mpOffscreenView;
}

BRect
IntervalSlider::BarFrame(void) const
{
	// Compute text height.
	const float lineHeight = mFontHeight.ascent
								  + mFontHeight.descent
								  + mFontHeight.leading;
	
	// Compute the bar's frame.
	BRect frame = Bounds();
	frame.InsetBy(kThumbWidth, 0.0);
	frame.top += lineHeight + kHashMarkSize;
	frame.bottom = frame.top + mBarThickness + 3.0;	// adjust for bevels

	return frame;
}

BRect
IntervalSlider::HashMarksFrame(void) const
{
	// Compute text height.
	const float lineHeight = mFontHeight.ascent
								  + mFontHeight.descent
								  + mFontHeight.leading;

	// Compute the hash mark frame.	
	BRect frame = Bounds();
	frame.InsetBy(kThumbWidth + 1.0, 0.0);	// adjust for bevels
	frame.top += lineHeight;
	frame.bottom = frame.top
					 + kHashMarkSize
					 + mBarThickness + 3.0		// see above
					 + kHashMarkSize;

	// Adjust the frame if only top or bottom hash marks are requested.
	switch (mHashMarks) {
	
	case B_HASH_MARKS_TOP:
		frame.bottom = frame.top + kHashMarkSize;
		break;
	
	case B_HASH_MARKS_BOTTOM:
		frame.top = frame.bottom - kHashMarkSize;
		break;
	
	default:
		// use complete frame
		break;
	}
	
	return frame;
}

BRect
IntervalSlider::LowThumbFrame(void) const
{
	// Get the bar's frame as reference.
	BRect frame = BarFrame();
	frame.left += 2;	// adjust for bevels
	--frame.right;

	// Compute horizontal range based on the low value.
	frame.right	= frame.left + frame.Width() * LowPosition();
	frame.left = frame.right - kThumbWidth;

	// Compute vertical range centered on the bar.	
	const float sum = frame.top + frame.bottom;
	frame.top = (sum - kThumbHeight) * 0.5;
	frame.bottom = (sum + kThumbHeight) * 0.5;
	
	return frame;
}

BRect
IntervalSlider::HighThumbFrame(void) const
{
	// Get the bar's frame as reference.
	BRect frame = BarFrame();
	frame.left += 2;	// adjust for bevels
	--frame.right;

	// Compute horizontal range based on the high value.
	frame.left = frame.left + frame.Width() * HighPosition();
	frame.right = frame.left + kThumbWidth;
	
	// Compute vertical range centered on the bar.	
	const float sum = frame.top + frame.bottom;
	frame.top = (sum - kThumbHeight) * 0.5;
	frame.bottom = (sum + kThumbHeight) * 0.5;
	
	return frame;
}

void
IntervalSlider::DrawSlider(void)
{
	// Erase offscreen view.
	mpOffscreenView->SetHighColor(mViewColor);
	mpOffscreenView->FillRect(mpOffscreenView->Bounds());

	// Draw individual components.
	DrawBar();
	DrawHashMarks();
	DrawThumbs();
	DrawText();
}

void
IntervalSlider::DrawBar(void)
{
	BRect frame = BarFrame();
	const rgb_color white = AdjustColor(kWhite);
	const rgb_color darkGray = AdjustColor(kDarkGray);
	const rgb_color black = AdjustColor(kBlack);
	
	// Draw bevel.
	mpOffscreenView->BeginLineArray(10);
	mpOffscreenView->AddLine(frame.LeftBottom(), frame.RightBottom(),white);
	mpOffscreenView->AddLine(frame.RightBottom(), frame.RightTop(),white);
	mpOffscreenView->AddLine(frame.LeftBottom(), frame.LeftTop(),darkGray);
	mpOffscreenView->AddLine(frame.LeftTop(), frame.RightTop(),darkGray);
	mpOffscreenView->AddLine(frame.LeftTop(), frame.LeftTop(), mViewColor);
	mpOffscreenView->AddLine(frame.RightTop(), frame.RightTop(), mViewColor);
	mpOffscreenView->AddLine(frame.LeftBottom(), frame.LeftBottom(), mViewColor);
	mpOffscreenView->AddLine(frame.RightBottom(), frame.RightBottom(), mViewColor);
	frame.InsetBy(1.0, 1.0);
	mpOffscreenView->AddLine(frame.LeftBottom(), frame.LeftTop(),black);
	mpOffscreenView->AddLine(frame.LeftTop(), frame.RightTop(),black);
	mpOffscreenView->EndLineArray();
	
	// Draw bar interior.
	++frame.left;
	++frame.top;
	const float lowX = frame.left + frame.Width() * LowPosition();
	const float highX = frame.left + frame.Width() * HighPosition();
	
	BRect lowFrame = frame;
	lowFrame.right = lowX;
	mpOffscreenView->SetHighColor(AdjustColor(mLowBarColor));
	mpOffscreenView->FillRect(lowFrame);
	
	BRect midFrame = frame;
	midFrame.left = lowX;
	midFrame.right = highX;
	mpOffscreenView->SetHighColor(AdjustColor(mMidBarColor));
	mpOffscreenView->FillRect(midFrame);
	
	BRect highFrame = frame;
	highFrame.left = highX;
	mpOffscreenView->SetHighColor(AdjustColor(mHighBarColor));
	mpOffscreenView->FillRect(highFrame);
}

void
IntervalSlider::DrawHashMarks(void)
{
	if (mHashMarks == B_HASH_MARKS_NONE || mHashMarkCount == 0) {
		return;
	}
	
	const BRect frame = HashMarksFrame();
	const float width = frame.Width();
	const rgb_color white = AdjustColor(kWhite);
	const rgb_color darkGray = AdjustColor(kDarkGray);

	if (mHashMarks != B_HASH_MARKS_BOTTOM) {
		// Draw top hash marks.
		mpOffscreenView->BeginLineArray(2 * mHashMarkCount);
		for (uint32 i = 0; i < mHashMarkCount; ++i) {
			const float x = frame.left + width * i / (mHashMarkCount - 1);
			mpOffscreenView->AddLine(BPoint(x, frame.top),
											 BPoint(x, frame.top + kHashMarkSize - 1),
											 darkGray);
			mpOffscreenView->AddLine(BPoint(x + 1, frame.top),
											 BPoint(x + 1, frame.top + kHashMarkSize - 1),
											 white);
		}
		mpOffscreenView->EndLineArray();
	}

	if (mHashMarks != B_HASH_MARKS_TOP) {
		// Draw bottom hash marks.
		mpOffscreenView->BeginLineArray(2 * mHashMarkCount);
		for (uint32 i = 0; i < mHashMarkCount; ++i) {
			const float x = frame.left + width * i / (mHashMarkCount - 1);
			mpOffscreenView->AddLine(BPoint(x, frame.bottom),
											 BPoint(x, frame.bottom - kHashMarkSize + 1),
											 darkGray);
			mpOffscreenView->AddLine(BPoint(x + 1, frame.bottom),
											 BPoint(x + 1, frame.bottom - kHashMarkSize + 1),
											 white);
		}
		mpOffscreenView->EndLineArray();
	}
}

void
IntervalSlider::DrawThumbs(void)
{
	mpOffscreenView->SetHighColor(AdjustColor(kLightGray));
	const rgb_color mediumGray = AdjustColor(kMediumGray);
	const rgb_color darkGray = AdjustColor(kDarkGray);
	const rgb_color black = AdjustColor(kBlack);

	// Draw low thumb...
	BRect frame = LowThumbFrame();
	frame.InsetBy(1.0, 1.0);
	--frame.bottom;
	// ...interior...
	mpOffscreenView->FillRect(frame);

	mpOffscreenView->BeginLineArray(8);
	// ...border...
	mpOffscreenView->AddLine(frame.LeftBottom() - BPoint(1.0, 0.0),
									 frame.LeftTop() - BPoint(1.0, 0.0),
									 darkGray);
	mpOffscreenView->AddLine(frame.LeftTop() - BPoint(0.0, 1.0),
									 frame.RightTop() - BPoint(0.0, 1.0),
									 darkGray);
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(0.0, 1.0),
									 frame.RightBottom() + BPoint(0.0, 1.0),
									 black);
	mpOffscreenView->AddLine(frame.RightBottom() + BPoint(1.0, 1.0),
									 frame.RightTop() + BPoint(1.0, -1.0),
									 black);
	// ...bevel...
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(1.0, 0.0),
									 frame.RightBottom(),
									 darkGray);
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(2.0, -1.0),
									 frame.RightBottom() - BPoint(0.0, 1.0),
									 mediumGray);
	// ...and shadow.
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(1.0, 2.0),
									 frame.RightBottom() + BPoint(1.0, 2.0),
									 darkGray);
	mpOffscreenView->AddLine(frame.RightBottom() + BPoint(2.0, 2.0),
									 frame.RightTop() + BPoint(2.0, 0.0),
									 darkGray);
	mpOffscreenView->EndLineArray();

	// Draw high thumb...
	frame = HighThumbFrame();
	frame.InsetBy(1.0, 1.0);
	--frame.bottom;
	// ...interior...
	mpOffscreenView->FillRect(frame);

	mpOffscreenView->BeginLineArray(11);
	// ...border...
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(-1.0, 1.0),
									 frame.LeftTop() - BPoint(1.0, 1.0),
									 black);
	mpOffscreenView->AddLine(frame.LeftTop() - BPoint(0.0, 1.0),
									 frame.RightTop() - BPoint(0.0, 1.0),
									 darkGray);
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(0.0, 1.0),
									 frame.RightBottom() + BPoint(0.0, 1.0),
									 black);
	mpOffscreenView->AddLine(frame.RightBottom() + BPoint(1.0, 0.0),
									 frame.RightTop() + BPoint(1.0, 0.0),
									 black);
	// ...bevel...
	mpOffscreenView->AddLine(frame.LeftBottom(),
									 frame.RightBottom(),
									 darkGray);
	mpOffscreenView->AddLine(frame.RightBottom(),
									 frame.RightTop() + BPoint(0.0, 1.0),
									 darkGray);
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(0.0, -1.0),
									 frame.RightBottom() - BPoint(1.0, 1.0),
									 mediumGray);
	mpOffscreenView->AddLine(frame.RightBottom() - BPoint(1.0, 1.0),
									 frame.RightTop() + BPoint(-1.0, 2.0),
									 mediumGray);
	// ...and shadow.
	mpOffscreenView->AddLine(frame.LeftBottom() + BPoint(0.0, 2.0),
									 frame.RightBottom() + BPoint(1.0, 2.0),
									 darkGray);
	mpOffscreenView->AddLine(frame.RightBottom() + BPoint(1.0, 1.0),
									 frame.RightBottom() + BPoint(1.0, 1.0),
									 darkGray);
	mpOffscreenView->AddLine(frame.RightBottom() + BPoint(2.0, 1.0),
									 frame.RightTop() + BPoint(2.0, 1.0),
									 darkGray);
	mpOffscreenView->EndLineArray();
}

void
IntervalSlider::DrawText(void)
{
	mpOffscreenView->SetHighColor(AdjustColor(kBlack));
	
	// Draw main label.	
	const BRect frame = Bounds();
	const BPoint labelPos = frame.LeftTop() + BPoint(0.0, mFontHeight.ascent);
	mpOffscreenView->DrawString(Label(), labelPos);
	
	// Draw limit labels.
	if (mpMinLabel) {
		const BPoint minLabelPos =
			frame.LeftBottom() - BPoint(0.0, mFontHeight.descent);
		mpOffscreenView->DrawString(mpMinLabel, minLabelPos);
	}
	
	if (mpMaxLabel) {
		const BPoint maxLabelPos =
			frame.RightBottom()
			- BPoint(mpOffscreenView->StringWidth(mpMaxLabel),
					   mFontHeight.descent);
		mpOffscreenView->DrawString(mpMaxLabel, maxLabelPos);
	}
	
	// Draw status text.
	if (const char *pText = UpdateText()) {
		const BPoint textPos =
			frame.RightTop()
			- BPoint(mpOffscreenView->StringWidth(pText), -mFontHeight.ascent);
		mpOffscreenView->DrawString(pText, textPos);
	}
}

const char *
IntervalSlider::UpdateText(void) const
{
	// No status message by default.
	return 0;
}

int16
IntervalSlider::ClampValue(int16 value) const
{
	if (value < mMinValue) {
		return mMinValue;
	} else if (value > mMaxValue) {
		return mMaxValue;
	}
	return value;
}

float
IntervalSlider::ValueToPosition(int16 value) const
{
	return (float)(value - mMinValue) / (float)(mMaxValue - mMinValue);
}

int16
IntervalSlider::PositionToValue(float position) const
{
	return mMinValue + (int16)(position * (mMaxValue - mMinValue));
}

rgb_color
IntervalSlider::AdjustColor(rgb_color color) const
{
	if (!IsEnabled()) {
		return tint_color(color, B_LIGHTEN_1_TINT);
	}
	return color;
}

void
IntervalSlider::CreateOffscreenView(void)
{
	if (mpOffscreenBits || mpOffscreenView) {
		debugger("Internal problem - offscreen view creation.");
	}

	BRect bounds = Bounds();
	mpOffscreenBits = new BBitmap(bounds, B_CMAP8, true);
	mpOffscreenView =
		new BView(bounds, "IntervalSlider offscreen", B_FOLLOW_NONE, B_WILL_DRAW);
	mpOffscreenBits->AddChild(mpOffscreenView);
}

void
IntervalSlider::DestroyOffscreenView(void)
{
	delete mpOffscreenBits;	// this also destroys the attached view
	mpOffscreenBits = 0;
	mpOffscreenView = 0;
}
