//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		A BSlider-like control with two knobs.

#ifndef INTERVALSLIDER_H
#define INTERVALSLIDER_H

class BBitmap;
#include <Control.h>
#include <Font.h>
#include <GraphicsDefs.h>
class BMessage;
#include <Point.h>
#include <Rect.h>
#include <Slider.h>
#include <SupportDefs.h>
class BView;

class IntervalSlider : public BControl
{
	typedef BControl inherited;
public:
	// Constructor and destructor.
	IntervalSlider(BRect frame, const char *pName, const char *pLabel,
						BMessage *pMsg, int16 minValue, int16 maxValue,
						uint32 resizingMode = B_FOLLOW_LEFT|B_FOLLOW_TOP,
						uint32 flags = 0);
	virtual ~IntervalSlider(void);

	// Hook functions.
	virtual void AttachedToWindow(void);
	virtual void DetachedFromWindow(void);

	virtual void Draw(BRect updateRect);

	virtual status_t Invoke(BMessage *pMsg = 0);
	virtual void SetModificationMessage(BMessage *pMsg);
	BMessage *ModificationMessage(void) const;

	virtual void FrameResized(float width, float height);
	virtual void GetPreferredSize( float *pWidth, float *pHeight);
	
	virtual void MouseDown(BPoint where);
	virtual void MouseUp(BPoint where);
	virtual void MouseMoved(BPoint where, uint32 code, const BMessage *pMsg);

//	virtual BHandler *ResolveSpecifier(BMessage *pMsg, int32 index,
//												  BMessage *pSpecifier, int32 form,
//												  const char *pProperty);
//	virtual status_t GetSupportedSuites(BMessage *pData);
	
	// Get limits of legal value range.
	void GetLimits(int32 &minimum, int32 &maximum) const;
	
	// The 32-bit BControl value is split into 2 16-bit halves
	// which are controlled separately by the two knobs.
	// There are convenience functions for setting and inquiring
	// both sub-values.
	virtual void SetValue(int32 value);
	void SetLowValue(int16 value);
	int16 LowValue(void) const;
	void SetHighValue(int16 value);
	int16 HighValue(void) const;
	void SetLowPosition(float Position);
	float LowPosition(void) const;
	void SetHighPosition(float Position);
	float HighPosition(void) const;

	// Management of visual details.
	virtual void SetLimitLabels(const char *pMinLabel, const char *pMaxLabel);
	const char *MinLimitLabel(void) const;
	const char *MaxLimitLabel(void) const;

	virtual void SetHashMarks(hash_mark_location where);
	hash_mark_location HashMarks() const;
	virtual void SetHashMarkCount(uint32 count);
	uint32 HashMarkCount(void) const;

	virtual void SetBarThickness(float thickness);
	float BarThickness(void) const;

	virtual void SetBarColors(rgb_color low, rgb_color mid, rgb_color high);
	rgb_color LowBarColor(void) const;
	rgb_color MidBarColor(void) const;
	rgb_color HighBarColor(void) const;

	virtual void SetFont(const BFont *pFont, uint32 properties = B_FONT_ALL);
	virtual void SetViewColor(rgb_color viewColor);
	void SetViewColor(uchar r, uchar g, uchar b, uchar a = 255)
	{
		rgb_color viewColor;
		viewColor.red = r;
		viewColor.green = g;
		viewColor.blue = b;
		viewColor.alpha = a;
		SetViewColor(viewColor);
	}

protected:
	// Hook functions for derived classes.
	virtual int16 ValueForPoint(BPoint where) const;

	BView &OffscreenView(void) const;
	virtual BRect BarFrame(void) const;
	virtual BRect HashMarksFrame(void) const;
	virtual BRect LowThumbFrame(void) const;
	virtual BRect HighThumbFrame(void) const;

	virtual void DrawSlider(void);
	virtual void DrawBar(void);
	virtual void DrawHashMarks(void);
	virtual void DrawThumbs(void);
	virtual void DrawText(void);
	virtual const char *UpdateText(void) const;
	
private:
	IntervalSlider(const IntervalSlider &);
	IntervalSlider &operator=(const IntervalSlider &);

	// Create and dispose of the offscreen view and bitmap.
	void CreateOffscreenView(void);
	void DestroyOffscreenView(void);

	// Clamp value to minimum/maximum.
	int16 ClampValue(int16 value) const;

	// Convert between values and relative positions.
	float ValueToPosition(int16 value) const;
	int16 PositionToValue(float position) const;

	// Dim colors if control is disabled.
	rgb_color AdjustColor(const rgb_color color) const;

	// Data members.	
	const int16 mMinValue;					// limits of the slider's value range
	const int16 mMaxValue;
	
	BMessage *mpModificationMessage;		// message sent while dragging
	enum {										// mouse tracking data
		NotTracking,
		LowTracking,
		HighTracking
	} mTracking;
	BPoint mTrackingOffset;

	char *mpMinLabel;							// limit labels
	char *mpMaxLabel;
	hash_mark_location mHashMarks;		// hash marks
	uint32 mHashMarkCount;
	float mBarThickness;						// bar thickness and colors
	rgb_color mLowBarColor;
	rgb_color mMidBarColor;
	rgb_color mHighBarColor;
	
	BBitmap *mpOffscreenBits;				// offscreen buffer
	BView *mpOffscreenView;
	rgb_color mViewColor;					// cached view color and font
	BFont mFont;
	font_height mFontHeight;
};

#endif // INTERVALSLIDER_H
