//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Configuration view for knot preferences.

#include "SamhainKnotConfigView.h"

#include <GraphicsDefs.h>
class BHandler;
#include <Message.h>
#include <Rect.h>
#include <Slider.h>
#include <SupportDefs.h>

#include "IntervalSlider.h"
#include "SamhainConfigViewBase.h"
#include "SamhainPreferences.h"

// Constant data.
static const float kProbabilityScale = 0.0001;		// 0.01% per slider tick
static const int16 kMinProbability = 0;
static const int16 kMaxProbability = 10000;
static const uint32 kHashMarksProbability = 11;

const rgb_color kBarColor  = { 128, 128, 255, 255 };
const rgb_color kFillColor = { 255, 128, 128, 255 };

static const char kFieldValue[]		= "be:value";
static const char kFieldLowValue[]	= "jjk:low_value";
static const char kFieldHighValue[]	= "jjk:high_value";

// Function object class for setting a single probability.
class SetProbabilityCommand : public SamhainConfigViewBase::Command
{
	typedef void (SamhainKnotPreferences::*SetProbability)(float);
public:
	SetProbabilityCommand(SamhainKnotPreferences &preferences,
								 const SetProbability pSetProbability)
	:	mPreferences(preferences),
		mpSetProbability(pSetProbability)
	{
		// empty
	}
	
	virtual void operator ()(BHandler &, BMessage &msg)
	{
		int32 value;
		if (msg.FindInt32(kFieldValue, &value) == B_OK) {
			(mPreferences.*mpSetProbability)(value * kProbabilityScale);
		}
	}

private:
	SetProbabilityCommand(const SetProbabilityCommand &);
	SetProbabilityCommand &operator =(const SetProbabilityCommand &);
	
	// Data members.
	SamhainKnotPreferences &mPreferences;
	const SetProbability mpSetProbability;
};

// Function object class for setting a probability range.
class SetDistributionCommand : public SamhainConfigViewBase::Command
{
	typedef void (SamhainKnotPreferences::*SetProbability)(float);
public:
	SetDistributionCommand(SamhainKnotPreferences &preferences,
								 const SetProbability pSetMinProbability,
								 const SetProbability pSetMaxProbability)
	:	mPreferences(preferences),
		mpSetMinProbability(pSetMinProbability),
		mpSetMaxProbability(pSetMaxProbability)
	{
		// empty
	}
	
	virtual void operator ()(BHandler &, BMessage &msg)
	{
		int16 value;
		if (msg.FindInt16(kFieldLowValue, &value) == B_OK) {
			(mPreferences.*mpSetMinProbability)(value * kProbabilityScale);
		}
		if (msg.FindInt16(kFieldHighValue, &value) == B_OK) {
			(mPreferences.*mpSetMaxProbability)(value * kProbabilityScale);
		}
	}

private:
	SetDistributionCommand(const SetDistributionCommand &);
	SetDistributionCommand &operator =(const SetDistributionCommand &);
	
	// Data members.
	SamhainKnotPreferences &mPreferences;
	const SetProbability mpSetMinProbability;
	const SetProbability mpSetMaxProbability;
};

// Configuration view.
SamhainKnotConfigView::SamhainKnotConfigView(
	BRect frame, const char *pName, SamhainKnotPreferences &preferences)
:	inherited(frame, pName),
	mPreferences(preferences),
	mpSetKnotDistribution(
		new SetDistributionCommand(preferences,
											&SamhainKnotPreferences::SetKnotProbability1,
											&SamhainKnotPreferences::SetKnotProbability2)),
	mpSetSectionDistribution(
		new SetDistributionCommand(preferences,
											&SamhainKnotPreferences::SetSectionProbability1,
											&SamhainKnotPreferences::SetSectionProbability2)),
	mpSetCornerDistribution(
		new SetDistributionCommand(preferences,
											&SamhainKnotPreferences::SetCornerProbability1,
											&SamhainKnotPreferences::SetCornerProbability2)),
	mpSetFatKnotProbability(
		new SetProbabilityCommand(preferences,
										  &SamhainKnotPreferences::SetFatKnotProbability)),
	mpSetHollowKnotProbability(
		new SetProbabilityCommand(preferences,
										  &SamhainKnotPreferences::SetHollowKnotProbability)),
	mpKnotSlider(0),
	mpSectionSlider(0),
	mpCornerSlider(0),
	mpFatKnotSlider(0),
	mpHollowKnotSlider(0)
{
	// empty
}

SamhainKnotConfigView::~SamhainKnotConfigView(void)
{
	delete mpSetKnotDistribution;
	mpSetKnotDistribution = 0;
	delete mpSetSectionDistribution;
	mpSetSectionDistribution = 0;
	delete mpSetCornerDistribution;
	mpSetCornerDistribution = 0;
	delete mpSetFatKnotProbability;
	mpSetFatKnotProbability = 0;
	delete mpSetHollowKnotProbability;
	mpSetHollowKnotProbability = 0;

	mpKnotSlider = 0;
	mpSectionSlider = 0;
	mpCornerSlider = 0;
	mpFatKnotSlider = 0;
	mpHollowKnotSlider = 0;
}

void
SamhainKnotConfigView::AllAttached(void)
{
	inherited::AllAttached();

	// Set/update control values.
	mpKnotSlider->SetLowValue((int16)(mPreferences.KnotProbability1() / kProbabilityScale));
	mpKnotSlider->SetHighValue((int16)(mPreferences.KnotProbability2() / kProbabilityScale));

	mpSectionSlider->SetLowValue((int16)(mPreferences.SectionProbability1() / kProbabilityScale));
	mpSectionSlider->SetHighValue((int16)(mPreferences.SectionProbability2() / kProbabilityScale));

	mpCornerSlider->SetLowValue((int16)(mPreferences.CornerProbability1() / kProbabilityScale));
	mpCornerSlider->SetHighValue((int16)(mPreferences.CornerProbability2() / kProbabilityScale));

	mpFatKnotSlider->SetValue((int32)(mPreferences.FatKnotProbability() / kProbabilityScale));

	mpHollowKnotSlider->SetValue((int32)(mPreferences.HollowKnotProbability() / kProbabilityScale));
}

void
SamhainKnotConfigView::AttachedToWindow(void)
{
	inherited::AttachedToWindow();

	if (!mpKnotSlider) {
		// Create the controls.
		BRect frame = Bounds();
		
		mpKnotSlider =
			new IntervalSlider(frame, "knots", "Knot types (H/C/V)",
									 CreateCommandMessage(*mpSetKnotDistribution),
									 kMinProbability, kMaxProbability);
		mpKnotSlider->SetLimitLabels("0%", "100%");
		mpKnotSlider->SetBarColors(kBarColor, kFillColor, kBarColor);
		mpKnotSlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpKnotSlider->SetHashMarkCount(kHashMarksProbability);
	
		AddChild(mpKnotSlider);
		mpKnotSlider->SetTarget(this);
		mpKnotSlider->ResizeToPreferred();
		frame.top += mpKnotSlider->Bounds().Height();
		--frame.top; // ### Save some space - for now.	
		
		mpSectionSlider =
			new IntervalSlider(frame, "sections", "Section types (H/D/V)",
									 CreateCommandMessage(*mpSetSectionDistribution),
									 kMinProbability, kMaxProbability);
		mpSectionSlider->SetLimitLabels("0%", "100%");
		mpSectionSlider->SetBarColors(kBarColor, kFillColor, kBarColor);
		mpSectionSlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpSectionSlider->SetHashMarkCount(kHashMarksProbability);
	
		AddChild(mpSectionSlider);
		mpSectionSlider->SetTarget(this);
		mpSectionSlider->ResizeToPreferred();
		frame.top += mpSectionSlider->Bounds().Height();
		--frame.top; // ### Save some space - for now.	
		
		mpCornerSlider =
			new IntervalSlider(frame, "corners", "Corner types (H/D/V)",
									 CreateCommandMessage(*mpSetCornerDistribution),
									 kMinProbability, kMaxProbability);
		mpCornerSlider->SetLimitLabels("0%", "100%");
		mpCornerSlider->SetBarColors(kBarColor, kFillColor, kBarColor);
		mpCornerSlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpCornerSlider->SetHashMarkCount(kHashMarksProbability);
	
		AddChild(mpCornerSlider);
		mpCornerSlider->SetTarget(this);
		mpCornerSlider->ResizeToPreferred();
		frame.top += mpCornerSlider->Bounds().Height();
		--frame.top; // ### Save some space - for now.	
		
		mpFatKnotSlider =
			new BSlider(frame, "fat knots", "Fat knots",
									 CreateCommandMessage(*mpSetFatKnotProbability),
									 kMinProbability, kMaxProbability);
		mpFatKnotSlider->SetLimitLabels("0%", "100%");
		mpFatKnotSlider->SetBarColor(kBarColor);
		mpFatKnotSlider->UseFillColor(true, &kFillColor);
		mpFatKnotSlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpFatKnotSlider->SetHashMarkCount(kHashMarksProbability);
	
		AddChild(mpFatKnotSlider);
		mpFatKnotSlider->SetTarget(this);
		mpFatKnotSlider->ResizeToPreferred();
		frame.top += mpFatKnotSlider->Bounds().Height();
		--frame.top; // ### Save some space - for now.	
		
		mpHollowKnotSlider =
			new BSlider(frame, "hollow knots", "Hollow knots",
									 CreateCommandMessage(*mpSetHollowKnotProbability),
									 kMinProbability, kMaxProbability);
		mpHollowKnotSlider->SetLimitLabels("0%", "100%");
		mpHollowKnotSlider->SetBarColor(kBarColor);
		mpHollowKnotSlider->UseFillColor(true, &kFillColor);
		mpHollowKnotSlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpHollowKnotSlider->SetHashMarkCount(kHashMarksProbability);
	
		AddChild(mpHollowKnotSlider);
		mpHollowKnotSlider->SetTarget(this);
		mpHollowKnotSlider->ResizeToPreferred();
		frame.top += mpHollowKnotSlider->Bounds().Height();
		--frame.top; // ### Save some space - for now.
	}
}
