//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Objects storing the screen saver preferences.

#include "SamhainPreferences.h"

#include <Autolock.h>
#include <Locker.h>
#include <Message.h>
#include <SupportDefs.h>

// Message fields.
const char kFieldMinNumKnots[] = "Samhain:MinNumKnots";
const char kFieldMaxNumKnots[] = "Samhain:MaxNumKnots";
const char kFieldMinDelay[] = "Samhain:MinDelay";
const char kFieldMaxDelay[] = "Samhain:MaxDelay";
const char kFieldKnotProbability1[] = "Samhain:KnotProbability1";
const char kFieldKnotProbability2[] = "Samhain:KnotProbability2";
const char kFieldSectionProbability1[] = "Samhain:SectionProbability1";
const char kFieldSectionProbability2[] = "Samhain:SectionProbability2";
const char kFieldCornerProbability1[] = "Samhain:CornerProbability1";
const char kFieldCornerProbability2[] = "Samhain:CornerProbability2";
const char kFieldFatKnotProbability[] = "Samhain:FatKnotProbability";
const char kFieldHollowKnotProbability[] = "Samhain:HollowKnotProbability";
const char kFieldMinSize[] = "Samhain:MinSize";
const char kFieldMaxSize[] = "Samhain:MaxSize";
const char kFieldSaverPreferences[] = "Samhain:SaverPreferences";
const char kFieldPreviewPreferences[] = "Samhain:PreviewPreferences";
const char kFieldKnotPreferences[] = "Samhain:KnotPreferences";

// Preferences for the screen saver.
SamhainSaverPreferences::SamhainSaverPreferences(BLocker &locker)
:	mLocker(locker)
{
	// empty
}

// Loading and saving the current state.
status_t
SamhainSaverPreferences::LoadState(BMessage &msg)
{
	BAutolock locker(mLocker);
	if (!locker.IsLocked()) {
		return B_BUSY;
	}
	
	status_t status;
	(void)
	(		(status = msg.FindInt16(kFieldMinNumKnots, &mMinNumKnots)) == B_OK
		&& (status = msg.FindInt16(kFieldMaxNumKnots, &mMaxNumKnots)) == B_OK
		&& (status = msg.FindInt64(kFieldMinDelay, &mMinDelay)) == B_OK
		&& (status = msg.FindInt64(kFieldMaxDelay, &mMaxDelay)) == B_OK
		&& (status = msg.FindFloat(kFieldMinSize, &mMinSize)) == B_OK
		&& (status = msg.FindFloat(kFieldMaxSize, &mMaxSize)) == B_OK);
	return status;
}

status_t
SamhainSaverPreferences::SaveState(BMessage &msg) const
{
	BAutolock locker(mLocker);
	if (!locker.IsLocked()) {
		return B_BUSY;
	}
	
	status_t status;
	(void)
	(		(status = msg.AddInt16(kFieldMinNumKnots, mMinNumKnots)) == B_OK
		&& (status = msg.AddInt16(kFieldMaxNumKnots, mMaxNumKnots)) == B_OK
		&& (status = msg.AddInt64(kFieldMinDelay, mMinDelay)) == B_OK
		&& (status = msg.AddInt64(kFieldMaxDelay, mMaxDelay)) == B_OK
		&& (status = msg.AddFloat(kFieldMinSize, mMinSize)) == B_OK
		&& (status = msg.AddFloat(kFieldMaxSize, mMaxSize)) == B_OK);
	return status;
}

// Inquiry methods.
int16
SamhainSaverPreferences::MinNumKnots(void) const
{
	BAutolock locker(mLocker);
	return mMinNumKnots;
}

int16
SamhainSaverPreferences::MaxNumKnots(void) const
{
	BAutolock locker(mLocker);
	return mMaxNumKnots;
}

bigtime_t
SamhainSaverPreferences::MinDelay(void) const
{
	BAutolock locker(mLocker);
	return mMinDelay;
}

bigtime_t
SamhainSaverPreferences::MaxDelay(void) const
{
	BAutolock locker(mLocker);
	return mMaxDelay;
}

float
SamhainSaverPreferences::MinSize(void) const
{
	BAutolock locker(mLocker);
	return mMinSize;
}

float
SamhainSaverPreferences::MaxSize(void) const
{
	BAutolock locker(mLocker);
	return mMaxSize;
}

// Modification methods.
void
SamhainSaverPreferences::SetMinNumKnots(int16 minNumKnots)
{
	BAutolock locker(mLocker);
	mMinNumKnots = minNumKnots;
}

void
SamhainSaverPreferences::SetMaxNumKnots(int16 maxNumKnots)
{
	BAutolock locker(mLocker);
	mMaxNumKnots = maxNumKnots;
}

void
SamhainSaverPreferences::SetMinDelay(bigtime_t minDelay)
{
	BAutolock locker(mLocker);
	mMinDelay = minDelay;
}

void
SamhainSaverPreferences::SetMaxDelay(bigtime_t maxDelay)
{
	BAutolock locker(mLocker);
	mMaxDelay = maxDelay;
}

void
SamhainSaverPreferences::SetMinSize(float minSize)
{
	BAutolock locker(mLocker);
	mMinSize = minSize;
}

void
SamhainSaverPreferences::SetMaxSize(float maxSize)
{
	BAutolock locker(mLocker);
	mMaxSize = maxSize;
}

// Preferences for the knot engine.
SamhainKnotPreferences::SamhainKnotPreferences(BLocker &locker)
:	mLocker(locker)
{
	// empty
}

// Loading and saving the current state.
status_t
SamhainKnotPreferences::LoadState(BMessage &msg)
{
	BAutolock locker(mLocker);
	if (!locker.IsLocked()) {
		return B_BUSY;
	}
	
	status_t status;
	(void)
	(		(status = msg.FindFloat(kFieldKnotProbability1, &mKnotProbability1)) == B_OK
		&& (status = msg.FindFloat(kFieldKnotProbability2, &mKnotProbability2)) == B_OK
		&& (status = msg.FindFloat(kFieldSectionProbability1, &mSectionProbability1)) == B_OK
		&& (status = msg.FindFloat(kFieldSectionProbability2, &mSectionProbability2)) == B_OK
		&& (status = msg.FindFloat(kFieldCornerProbability1, &mCornerProbability1)) == B_OK
		&& (status = msg.FindFloat(kFieldCornerProbability2, &mCornerProbability2)) == B_OK
		&& (status = msg.FindFloat(kFieldFatKnotProbability, &mFatKnotProbability)) == B_OK
		&& (status = msg.FindFloat(kFieldHollowKnotProbability, &mHollowKnotProbability)) == B_OK);
	return status;
}

status_t
SamhainKnotPreferences::SaveState(BMessage &msg) const
{
	BAutolock locker(mLocker);
	if (!locker.IsLocked()) {
		return B_BUSY;
	}
	
	status_t status;
	(void)
	(		(status = msg.AddFloat(kFieldKnotProbability1, mKnotProbability1)) == B_OK
		&& (status = msg.AddFloat(kFieldKnotProbability2, mKnotProbability2)) == B_OK
		&& (status = msg.AddFloat(kFieldSectionProbability1, mSectionProbability1)) == B_OK
		&& (status = msg.AddFloat(kFieldSectionProbability2, mSectionProbability2)) == B_OK
		&& (status = msg.AddFloat(kFieldCornerProbability1, mCornerProbability1)) == B_OK
		&& (status = msg.AddFloat(kFieldCornerProbability2, mCornerProbability2)) == B_OK
		&& (status = msg.AddFloat(kFieldFatKnotProbability, mFatKnotProbability)) == B_OK
		&& (status = msg.AddFloat(kFieldHollowKnotProbability, mHollowKnotProbability)) == B_OK);
	return status;
}

// Inquiry methods.
float
SamhainKnotPreferences::KnotProbability1(void) const
{
	BAutolock locker(mLocker);
	return mKnotProbability1;
}

float
SamhainKnotPreferences::KnotProbability2(void) const
{
	BAutolock locker(mLocker);
	return mKnotProbability2;
}

float
SamhainKnotPreferences::SectionProbability1(void) const
{
	BAutolock locker(mLocker);
	return mSectionProbability1;
}

float
SamhainKnotPreferences::SectionProbability2(void) const
{
	BAutolock locker(mLocker);
	return mSectionProbability2;
}

float
SamhainKnotPreferences::CornerProbability1(void) const
{
	BAutolock locker(mLocker);
	return mCornerProbability1;
}

float
SamhainKnotPreferences::CornerProbability2(void) const
{
	BAutolock locker(mLocker);
	return mCornerProbability2;
}

float
SamhainKnotPreferences::FatKnotProbability(void) const
{
	BAutolock locker(mLocker);
	return mFatKnotProbability;
}

float
SamhainKnotPreferences::HollowKnotProbability(void) const
{
	BAutolock locker(mLocker);
	return mHollowKnotProbability;
}

// Modification methods.
void
SamhainKnotPreferences::SetKnotProbability1(float knotProbability1)
{
	BAutolock locker(mLocker);
	mKnotProbability1 = knotProbability1;
}

void
SamhainKnotPreferences::SetKnotProbability2(float knotProbability2)
{
	BAutolock locker(mLocker);
	mKnotProbability2 = knotProbability2;
}

void
SamhainKnotPreferences::SetSectionProbability1(float sectionProbability1)
{
	BAutolock locker(mLocker);
	mSectionProbability1 = sectionProbability1;
}

void
SamhainKnotPreferences::SetSectionProbability2(float sectionProbability2)
{
	BAutolock locker(mLocker);
	mSectionProbability2 = sectionProbability2;
}

void
SamhainKnotPreferences::SetCornerProbability1(float cornerProbability1)
{
	BAutolock locker(mLocker);
	mCornerProbability1 = cornerProbability1;
}

void
SamhainKnotPreferences::SetCornerProbability2(float cornerProbability2)
{
	BAutolock locker(mLocker);
	mCornerProbability2 = cornerProbability2;
}

void
SamhainKnotPreferences::SetFatKnotProbability(float fatKnotProbability)
{
	BAutolock locker(mLocker);
	mFatKnotProbability = fatKnotProbability;
}

void
SamhainKnotPreferences::SetHollowKnotProbability(float hollowKnotProbability)
{
	BAutolock locker(mLocker);
	mHollowKnotProbability = hollowKnotProbability;
}

// All preferences.
SamhainPreferences::SamhainPreferences(void)
:	mLocker("SamhainPreferences"),
	mSaverPreferences(mLocker),
	mPreviewPreferences(mLocker),
	mKnotPreferences(mLocker)
{
	// empty
}

// Loading and saving the current state.
status_t
SamhainPreferences::LoadState(BMessage &msg)
{
	BAutolock locker(mLocker);
	if (!locker.IsLocked()) {
		return B_BUSY;
	}
	
	status_t status;
	{
		BMessage message;
		if ((status = msg.FindMessage(kFieldSaverPreferences, &message)) == B_OK) {
			status = mSaverPreferences.LoadState(message);
		}
	}

	if (status == B_OK) {
		BMessage message;
		if ((status = msg.FindMessage(kFieldPreviewPreferences, &message)) == B_OK) {
			status = mPreviewPreferences.LoadState(message);
		}
	}

	if (status == B_OK) {
		BMessage message;
		if ((status = msg.FindMessage(kFieldKnotPreferences, &message)) == B_OK) {
			status = mKnotPreferences.LoadState(message);
		}
	}

	return status;
}

status_t
SamhainPreferences::SaveState(BMessage &msg) const
{
	BAutolock locker(mLocker);
	if (!locker.IsLocked()) {
		return B_BUSY;
	}
	
	status_t status;
	{
		BMessage message;
		if ((status = mSaverPreferences.SaveState(message)) == B_OK) {
			status = msg.AddMessage(kFieldSaverPreferences, &message);
		}
	}
	
	if (status == B_OK) {
		BMessage message;
		if ((status = mPreviewPreferences.SaveState(message)) == B_OK) {
			status = msg.AddMessage(kFieldPreviewPreferences, &message);
		}
	}

	if (status == B_OK) {
		BMessage message;
		if ((status = mKnotPreferences.SaveState(message)) == B_OK) {
			status = msg.AddMessage(kFieldKnotPreferences, &message);
		}
	}
	
	return status;
}

// Access to components.
const SamhainSaverPreferences &
SamhainPreferences::SaverPreferences(void) const
{
	return mSaverPreferences;
}

SamhainSaverPreferences &
SamhainPreferences::SaverPreferences(void)
{
	return mSaverPreferences;
}

const SamhainSaverPreferences &
SamhainPreferences::PreviewPreferences(void) const
{
	return mPreviewPreferences;
}

SamhainSaverPreferences &
SamhainPreferences::PreviewPreferences(void)
{
	return mPreviewPreferences;
}

const SamhainKnotPreferences &
SamhainPreferences::KnotPreferences(void) const
{
	return mKnotPreferences;
}

SamhainKnotPreferences &
SamhainPreferences::KnotPreferences(void)
{
	return mKnotPreferences;
}

// Establish default preferences.
void
SamhainPreferences::SetDefaults(void)
{
	// 1–5 knots spaced 3–6 seconds in screen saving mode.
	mSaverPreferences.SetMinNumKnots(1);
	mSaverPreferences.SetMaxNumKnots(5);
	mSaverPreferences.SetMinDelay(3000000);
	mSaverPreferences.SetMaxDelay(6000000);
	mSaverPreferences.SetMinSize(0.2);
	mSaverPreferences.SetMaxSize(1.0);
	
	// 1–5 knots spaced 1 second in preview mode.
	mPreviewPreferences.SetMinNumKnots(1);
	mPreviewPreferences.SetMaxNumKnots(5);
	mPreviewPreferences.SetMinDelay(1000000);
	mPreviewPreferences.SetMaxDelay(1000000);
	mPreviewPreferences.SetMinSize(0.2);
	mPreviewPreferences.SetMaxSize(1.0);
	
	// 20% horizontal, 60% closed and 20% vertical knots.
	mKnotPreferences.SetKnotProbability1(0.2);
	mKnotPreferences.SetKnotProbability2(0.8);
	// 6% horizontal, 88% diagonal and 6% vertical sections.
	mKnotPreferences.SetSectionProbability1(0.06);
	mKnotPreferences.SetSectionProbability2(0.94);
	// 20% horizontal, 60% diagonal and 20% vertical corners.
	mKnotPreferences.SetCornerProbability1(0.2);
	mKnotPreferences.SetCornerProbability2(0.8);
	// 40% fat, 60% slim knots.
	mKnotPreferences.SetFatKnotProbability(0.4);
	// 50% hollow, 50% solid knots.
	mKnotPreferences.SetHollowKnotProbability(0.5);
}
