//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Configuration view for saver/preview preferences.

#include "SamhainSaverConfigView.h"

#include <cstdio>

#include <GraphicsDefs.h>
class BHandler;
#include <Message.h>
#include <Rect.h>
#include <SupportDefs.h>

#include "IntervalSlider.h"
#include "SamhainConfigViewBase.h"
#include "SamhainPreferences.h"

// Constant data.
static const bigtime_t kTimeScale = 10000;		// 10msec per delay slider tick
static const bigtime_t kSecondsScale = 1000000;

static const float kSizeScale = 0.0001;			// 0.01% per size slider tick
static const int16 kMinSize = 0;
static const int16 kMaxSize = 10000;
static const uint32 kHashMarksSize = 11;

const rgb_color kBarColor  = { 128, 128, 255, 255 };
const rgb_color kFillColor = { 255, 128, 128, 255 };

static const char kFieldLowValue[]  = "jjk:low_value";
static const char kFieldHighValue[] = "jjk:high_value";

// Function object class for setting number of knots.
class SetNumKnotsCommand : public SamhainConfigViewBase::Command
{
public:
	SetNumKnotsCommand(SamhainSaverPreferences &preferences)
	:	mPreferences(preferences)
	{
		// empty
	}
	
	virtual void operator ()(BHandler &, BMessage &msg)
	{
		int16 value;
		if (msg.FindInt16(kFieldLowValue, &value) == B_OK) {
			mPreferences.SetMinNumKnots(value);
		}
		if (msg.FindInt16(kFieldHighValue, &value) == B_OK) {
			mPreferences.SetMaxNumKnots(value);
		}
	}

private:
	SetNumKnotsCommand(const SetNumKnotsCommand &);
	SetNumKnotsCommand &operator =(const SetNumKnotsCommand &);
	
	// Data members.
	SamhainSaverPreferences &mPreferences;
};

// Function object class for setting delay between knots.
class SetDelayCommand : public SamhainConfigViewBase::Command
{
public:
	SetDelayCommand(SamhainSaverPreferences &preferences)
	:	mPreferences(preferences)
	{
		// empty
	}
	
	virtual void operator ()(BHandler &, BMessage &msg)
	{
		int16 value;
		if (msg.FindInt16(kFieldLowValue, &value) == B_OK) {
			mPreferences.SetMinDelay(value * kTimeScale);
		}
		if (msg.FindInt16(kFieldHighValue, &value) == B_OK) {
			mPreferences.SetMaxDelay(value * kTimeScale);
		}
	}

private:
	SetDelayCommand(const SetDelayCommand &);
	SetDelayCommand &operator =(const SetDelayCommand &);
	
	// Data members.
	SamhainSaverPreferences &mPreferences;
};

// Function object class for setting size of knots.
class SetSizeCommand : public SamhainConfigViewBase::Command
{
public:
	SetSizeCommand(SamhainSaverPreferences &preferences)
	:	mPreferences(preferences)
	{
		// empty
	}
	
	virtual void operator ()(BHandler &, BMessage &msg)
	{
		int16 value;
		if (msg.FindInt16(kFieldLowValue, &value) == B_OK) {
			mPreferences.SetMinSize(value * kSizeScale);
		}
		if (msg.FindInt16(kFieldHighValue, &value) == B_OK) {
			mPreferences.SetMaxSize(value * kSizeScale);
		}
	}

private:
	SetSizeCommand(const SetSizeCommand &);
	SetSizeCommand &operator =(const SetSizeCommand &);
	
	// Data members.
	SamhainSaverPreferences &mPreferences;
};

// Configuration view.
SamhainSaverConfigView::SamhainSaverConfigView(
	BRect frame, const char *pName,
	int16 minNumKnots, int16 maxNumKnots, uint32 hashMarksNumKnots,
	bigtime_t minDelay, bigtime_t maxDelay, uint32 hashMarksDelay,
	SamhainSaverPreferences &preferences)
:	inherited(frame, pName),
	mMinNumKnots(minNumKnots),
	mMaxNumKnots(maxNumKnots),
	mHashMarksNumKnots(hashMarksNumKnots),
	mMinDelay(minDelay),
	mMaxDelay(maxDelay),
	mHashMarksDelay(hashMarksDelay),
	mPreferences(preferences),
	mpSetNumKnots(new SetNumKnotsCommand(preferences)),
	mpSetDelay(new SetDelayCommand(preferences)),
	mpSetSize(new SetSizeCommand(preferences)),
	mpKnotsSlider(0),
	mpDelaySlider(0),
	mpSizeSlider(0)
{
	// empty
}

SamhainSaverConfigView::~SamhainSaverConfigView(void)
{
	delete mpSetNumKnots;
	mpSetNumKnots = 0;
	delete mpSetDelay;
	mpSetDelay = 0;
	delete mpSetSize;
	mpSetSize = 0;

	mpKnotsSlider = 0;
	mpDelaySlider = 0;
	mpSizeSlider = 0;
}

void
SamhainSaverConfigView::AllAttached(void)
{
	inherited::AllAttached();

	// Set/update control values.
	mpKnotsSlider->SetLowValue(mPreferences.MinNumKnots());
	mpKnotsSlider->SetHighValue(mPreferences.MaxNumKnots());
	
	mpDelaySlider->SetLowValue((int16)(mPreferences.MinDelay() / kTimeScale));
	mpDelaySlider->SetHighValue((int16)(mPreferences.MaxDelay() / kTimeScale));
	
	mpSizeSlider->SetLowValue((int16)(mPreferences.MinSize() / kSizeScale));
	mpSizeSlider->SetHighValue((int16)(mPreferences.MaxSize() / kSizeScale));
}

void
SamhainSaverConfigView::AttachedToWindow(void)
{
	inherited::AttachedToWindow();
	
	if (!mpKnotsSlider) {
		// Create the controls.
		BRect frame = Bounds();
		char minLimitLabel[100], maxLimitLabel[100];
		
		mpKnotsSlider =
			new IntervalSlider(frame, "knots", "Number of knots",
									 CreateCommandMessage(*mpSetNumKnots),
									 mMinNumKnots, mMaxNumKnots);
		sprintf(minLimitLabel, "%d", mMinNumKnots);
		sprintf(maxLimitLabel, "%d", mMaxNumKnots);
		mpKnotsSlider->SetLimitLabels(minLimitLabel, maxLimitLabel);
		mpKnotsSlider->SetBarColors(kBarColor, kFillColor, kBarColor);
		mpKnotsSlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpKnotsSlider->SetHashMarkCount(mHashMarksNumKnots);
	
		AddChild(mpKnotsSlider);
		mpKnotsSlider->SetTarget(this);
		mpKnotsSlider->ResizeToPreferred();
		frame.top += mpKnotsSlider->Bounds().Height();
		
		mpDelaySlider =
			new IntervalSlider(frame, "delay", "Delay between knots",
									 CreateCommandMessage(*mpSetDelay),
									 (int16)(mMinDelay / kTimeScale),
									 (int16)(mMaxDelay / kTimeScale));
		sprintf(minLimitLabel, "%dsec", (int)(mMinDelay / kSecondsScale));
		sprintf(maxLimitLabel, "%dsec", (int)(mMaxDelay / kSecondsScale));
		mpDelaySlider->SetLimitLabels(minLimitLabel, maxLimitLabel);
		mpDelaySlider->SetBarColors(kBarColor, kFillColor, kBarColor);
		mpDelaySlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpDelaySlider->SetHashMarkCount(mHashMarksDelay);
	
		AddChild(mpDelaySlider);
		mpDelaySlider->SetTarget(this);
		mpDelaySlider->ResizeToPreferred();
		frame.top += mpDelaySlider->Bounds().Height();
	
		mpSizeSlider =
			new IntervalSlider(frame, "size", "Size of knots",
									 CreateCommandMessage(*mpSetSize),
									 kMinSize, kMaxSize);
		mpSizeSlider->SetLimitLabels("0%", "100%");
		mpSizeSlider->SetBarColors(kBarColor, kFillColor, kBarColor);
		mpSizeSlider->SetHashMarks(B_HASH_MARKS_BOTH);
		mpSizeSlider->SetHashMarkCount(kHashMarksSize);
	
		AddChild(mpSizeSlider);
		mpSizeSlider->SetTarget(this);
		mpSizeSlider->ResizeToPreferred();
		frame.top += mpSizeSlider->Bounds().Height();
		
		// ## The size control doesn't work yet.
		mpSizeSlider->SetEnabled(false);
	}
}
