//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Draw and cache the knot's sections.  See KnotSection.h for more info.

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <ctime>

#include <Bitmap.h>
#include <ByteOrder.h>
#include <Debug.h>
#include <GraphicsDefs.h>
#include <Point.h>
#include <Rect.h>
#include <Region.h>
#include <View.h>

#include "KnotSectionCache.h"

KnotSectionCache::KnotSectionCache(
	int32 edgeLength, float outlineWidth, float fillWidth,
	bool hollow, rgb_color color)
:	inherited(BRect(B_ORIGIN, BPoint(edgeLength * 4.0, edgeLength * 4.0)),
				 "KnotSectionCache",
				 B_FOLLOW_NONE,
				 B_WILL_DRAW|B_SUBPIXEL_PRECISE),
	mSize(edgeLength * 2.0),
	mCenter(mSize, mSize),
	mOutlineWidth(edgeLength * outlineWidth * 4.0),
	mFillWidth(edgeLength * fillWidth * 4.0),
	mHollow(hollow),
	mColor(color),
	mBuffer(BRect(B_ORIGIN, BPoint(edgeLength - 1, edgeLength - 1)), B_RGB32),
	mCanvas(BRect(B_ORIGIN, BPoint(edgeLength*4 - 1, edgeLength*4 - 1)),
			  B_GRAY1, true)
{
	// Initialize drawing modes.
	SetDrawingMode(B_OP_COPY);
	SetLineMode(B_BUTT_CAP, B_MITER_JOIN);
	
	// Clear the cache.
	memset(mpCache, 0, sizeof(BBitmap *) * 3*4*4*4*4);

	// Attach to the drawing bitmap.
	mCanvas.AddChild(this);
}

KnotSectionCache::~KnotSectionCache(void)
{
	// Detach from the drawing bitmap.
	mCanvas.RemoveChild(this);
	
	// Free cached sections.
	for (int i = 0; i < 3; ++i) {
		for (int j = 0; j < 4; ++j) {
			for (int k = 0; k < 4; ++k) {
				for (int l = 0; l < 4; ++l) {
					for (int m = 0; m < 4; ++m) {
						delete mpCache[i][j][k][l][m];
						mpCache[i][j][k][l][m] = 0;
					}
				}
			}
		}
	}
}

const BBitmap *
KnotSectionCache::GetKnotSection(const KnotSection &section)
{
	ASSERT(section.Type() != N);
	
	// Look up the section in the cache.
	BBitmap *&pCached =
		mpCache
			[section.Type()]
			[section.LeftTop()]
			[section.RightTop()]
			[section.LeftBottom()]
			[section.RightBottom()];
	
	if (!pCached) {
		// Draw the section.
		DrawKnotSection(section);
		AntiAlias();
	
		// Cache the resulting bitmap.
		pCached = new BBitmap(&mBuffer);
	}
	
	return pCached;
}

// Private methods for drawing the sections.
void
KnotSectionCache::DrawKnotSection(const KnotSection &section)
{
	if (mCanvas.Lock()) {
		// Erase the bounds rectangle.
		FillRect(Bounds(), B_SOLID_LOW);
		
		// Draw the section.
		switch (section.Type()) {
		
		case D:
			DiagonalSection(section);
			break;
		
		case H:
			HorizontalSection(section);
			break;
		
		case V:
			VerticalSection(section);
			break;
		
		default:
			DEBUGGER("Illegal section type.");
			break;
		}

		Sync();
		mCanvas.Unlock();
	}
}

void
KnotSectionCache::DiagonalSection(const KnotSection &section)
{
	// Draw bottom stroke.
	if (mHollow) {
		SetPenSize(mOutlineWidth);
		DiagonalStroke(section.RightTop(), section.LeftBottom(), true, false);
	}
	SetPenSize(mFillWidth);
	DiagonalStroke(section.RightTop(), section.LeftBottom(), true, mHollow);

	// Draw top stroke.
	SetPenSize(mOutlineWidth);
	DiagonalStroke(section.LeftTop(), section.RightBottom(), false, !mHollow);
	SetPenSize(mFillWidth);
	DiagonalStroke(section.LeftTop(), section.RightBottom(), false, mHollow);

	// Draw corners if needed.
	StrokeCorners(section);
}

void
KnotSectionCache::DiagonalStroke(
	direction from, direction to, bool mirror, bool erase)
{
	switch (4*from + to) {

	case 4*D + D:
		StrokeDDD(mirror ? 4 : 0, erase);
		break;
	
	case 4*D + H:
		StrokeDHD(mirror ? 6 : 2, erase);
		break;
	
	case 4*D + V:
		StrokeDHD(mirror ? 1 : 5, erase);
		break;
	
	case 4*H + D:
		StrokeDHD(mirror ? 4 : 0, erase);
		break;
	
	case 4*H + H:
		StrokeDHH(mirror ? 4 : 0, erase);
		break;
	
	case 4*H + V:
		StrokeDHV(mirror ? 4 : 0, erase);
		break;
	
	case 4*V + D:
		StrokeDHD(mirror ? 3 : 7, erase);
		break;
	
	case 4*V + H:
		StrokeDHV(mirror ? 6 : 2, erase);
		break;
	
	case 4*V + V:
		StrokeDHH(mirror ? 1 : 5, erase);
		break;	
	
	default:
		DEBUGGER("Illegal diagonal stroke.");
		break;
	}
}

void
KnotSectionCache::HorizontalSection(const KnotSection &section)
{
	// Draw bottom stroke.
	if (mHollow) {
		SetPenSize(mOutlineWidth);
		HorizontalStroke(section.LeftBottom(), section.RightBottom(),
							  true, false);
	}
	SetPenSize(mFillWidth);
	HorizontalStroke(section.LeftBottom(), section.RightBottom(),
						  true, mHollow);

	// Draw top stroke.
	if (mHollow) {
		SetPenSize(mOutlineWidth);
		HorizontalStroke(section.LeftTop(), section.RightTop(), false, false);
	}
	SetPenSize(mFillWidth);
	HorizontalStroke(section.LeftTop(), section.RightTop(), false, mHollow);

	// Draw corners if needed.
	StrokeCorners(section);
}

void
KnotSectionCache::HorizontalStroke(
	direction from, direction to, bool mirror, bool erase)
{
	switch (4*from + to) {

	case 4*D + D:
		StrokeHDD(mirror ? 2 : 0, erase);
		break;
	
	case 4*D + H:
		StrokeHDH(mirror ? 6 : 0, erase);
		break;
	
	case 4*D + V:
		StrokeHDV(mirror ? 6 : 0, erase);
		break;
	
	case 4*H + D:
		StrokeHDH(mirror ? 2 : 4, erase);
		break;
	
	case 4*H + H:
		StrokeHHH(mirror ? 2 : 0, erase);
		break;
	
	case 4*H + V:
		StrokeHHV(mirror ? 6 : 0, erase);
		break;
	
	case 4*V + D:
		StrokeHDV(mirror ? 2 : 4, erase);
		break;
	
	case 4*V + H:
		StrokeHHV(mirror ? 2 : 4, erase);
		break;
	
	case 4*V + V:
		StrokeHVV(mirror ? 2 : 0, erase);
		break;	
	
	case 4*N + N:
		// Stroke omitted (edge section).
		break;
	
	default:
		DEBUGGER("Illegal horizontal stroke.");
		break;
	}
}

void
KnotSectionCache::VerticalSection(const KnotSection &section)
{
	// Draw right stroke.
	if (mHollow) {
		SetPenSize(mOutlineWidth);
		VerticalStroke(section.RightTop(), section.RightBottom(), true, false);
	}
	SetPenSize(mFillWidth);
	VerticalStroke(section.RightTop(), section.RightBottom(), true, mHollow);

	// Draw left stroke.
	if (mHollow) {
		SetPenSize(mOutlineWidth);
		VerticalStroke(section.LeftTop(), section.LeftBottom(), false, false);
	}
	SetPenSize(mFillWidth);
	VerticalStroke(section.LeftTop(), section.LeftBottom(), false, mHollow);

	// Draw corners if needed.
	StrokeCorners(section);
}

void
KnotSectionCache::VerticalStroke(
	direction from, direction to, bool mirror, bool erase)
{
	switch (4*from + to) {
	
	case 4*D + D:
		StrokeHDD(mirror ? 3 : 1, erase);
		break;
	
	case 4*D + H:
		StrokeHDV(mirror ? 3 : 7, erase);
		break;
	
	case 4*D + V:
		StrokeHDH(mirror ? 3 : 7, erase);
		break;
	
	case 4*H + D:
		StrokeHDV(mirror ? 5 : 1, erase);
		break;
	
	case 4*H + H:
		StrokeHVV(mirror ? 3 : 1, erase);
		break;
	
	case 4*H + V:
		StrokeHHV(mirror ? 5 : 1, erase);
		break;
	
	case 4*V + D:
		StrokeHDH(mirror ? 5 : 1, erase);
		break;
	
	case 4*V + H:
		StrokeHHV(mirror ? 3 : 7, erase);
		break;
	
	case 4*V + V:
		StrokeHHH(mirror ? 3 : 1, erase);
		break;	
	
	case 4*N + N:
		// Stroke omitted (edge section).
		break;
	
	default:
		DEBUGGER("Illegal vertical stroke.");
		break;
	}
}

void
KnotSectionCache::StrokeCorners(const KnotSection &section)
{
	// Draw corners of the section if needed.
	if (section.LeftTop() == D) {
		SetPenSize(mOutlineWidth);
		StrokeLeftTop(!mHollow);
		SetPenSize(mFillWidth);
		StrokeLeftTop(mHollow);
	}
	if (section.RightBottom() == D) {
		SetPenSize(mOutlineWidth);
		StrokeRightBottom(!mHollow);
		SetPenSize(mFillWidth);
		StrokeRightBottom(mHollow);
	}
}

void
KnotSectionCache::StrokeLeftTop(bool erase)
{
	// The top left crossing stroke of a diagonal.
	StrokeLine(mCenter - BPoint(2.0 * mSize, 0),
				  mCenter - BPoint(0, 2.0 * mSize),
				  erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeRightBottom(bool erase)
{
	// The bottom right crossing stroke of a diagonal.
	StrokeLine(mCenter + BPoint(2.0 * mSize, 0),
				  mCenter + BPoint(0, 2.0 * mSize),
				  erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

inline float
orientAngle(float angle, int orientation)
{
	switch (orientation) {
	case 1:
	case 5:
		angle += 90.0;
		break;

	case 2:
	case 6:
		angle += 180.0;
		break;

	case 3:
	case 7:
		angle += 270.0;
		break;

	default:
		break;
	}
	
	if (orientation > 3) {
		angle = 180.0 - angle;
	}

	return angle;
}

inline BPoint
orientPoint(float x, float y, int orientation)
{
	BPoint point(x, y);
	
	switch (orientation) {
	case 1:
	case 5:
		point.x = y;
		point.y = -x;
		break;

	case 2:
	case 6:
		point.x = -x;
		point.y = -y;
		break;

	case 3:
	case 7:
		point.x = -y;
		point.y = x;
		break;

	default:
		break;
	}
	
	if (orientation > 3) {
		point.x = -point.x;
	}
	
	return point;
}

void
KnotSectionCache::StrokeDDD(int orientation, bool erase)
{
	// The basic diagonal stroke.
	const BPoint offset = orientPoint(mSize, mSize, orientation);
	StrokeLine(mCenter - offset, mCenter + offset,
				  erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeDHD(int orientation, bool erase)
{
	// Diagonal stroke, horizontal to diagonal.
	const float radius = fabs(mSize) / 2.0 / (M_SQRT2 - 1.0);
	const BPoint arcOffset = orientPoint(-mSize, radius - mSize*0.5, orientation);
	const float angle = orientAngle(67.5, orientation);
	StrokeArc(mCenter + arcOffset, radius, radius, angle - 25.0, 50.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
	
	const float strokeOffset = mSize * (1.0 / 2.0 / (2.0 - M_SQRT2) - 1.0) - 1.0;
	StrokeLine(mCenter + orientPoint(strokeOffset, strokeOffset, orientation),
				  mCenter + orientPoint(mSize, mSize, orientation),
				  erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeDHH(int orientation, bool erase)
{
	// Diagonal stroke, horizontal to horizontal.
	// ### Temporary workaround until Be fixes StrokeBezier().
	const float radius = mSize * 1.25;
	StrokeArc(mCenter + orientPoint(-mSize, -mSize * 0.5 + radius, orientation),
				 radius, radius, orientAngle(64.0, orientation) - 27.0, 54.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
	StrokeArc(mCenter + orientPoint(mSize, mSize * 0.5 - radius, orientation),
				 radius, radius, orientAngle(244.0, orientation) - 27.0, 54.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeDHV(int orientation, bool erase)
{
	// Diagonal stroke, horizontal to vertical.
	const float radius = mSize * 1.5;
	StrokeArc(mCenter + orientPoint(-mSize, mSize, orientation),
				 radius, radius, orientAngle(45.0, orientation) - 48.0, 96.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeHDD(int orientation, bool erase)
{
	// Horizontal stroke, diagonal to diagonal.
	const BPoint radius =
		orientPoint(mSize, mSize * 0.75, orientation);
	const float xRadius = fabs(radius.x);
	const float yRadius = fabs(radius.y);
	StrokeArc(mCenter + orientPoint(0, -1.25 * mSize, orientation),
				 xRadius, yRadius, orientAngle(270.0, orientation) - 96.0, 192.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeHDH(int orientation, bool erase)
{
	// Horizontal stroke, diagonal to horizontal.
	const BPoint radius =
		orientPoint(mSize, mSize * 0.75, orientation);
	const float xRadius = fabs(radius.x);
	const float yRadius = fabs(radius.y);
	StrokeArc(mCenter + orientPoint(0, -1.25 * mSize, orientation),
				 xRadius, yRadius, orientAngle(225.0, orientation) - 48.0, 96.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);

	StrokeLine(mCenter + orientPoint(0, -0.5 * mSize, orientation),
				  mCenter + orientPoint(mSize, -0.5 * mSize, orientation),
				  erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeHDV(int orientation, bool erase)
{
	// Horizontal stroke, diagonal to vertical.
	const BPoint radius =
		orientPoint(mSize, mSize * 0.75, orientation);
	const float xRadius = fabs(radius.x);
	const float yRadius = fabs(radius.y);
	StrokeArc(mCenter + orientPoint(0, -1.25 * mSize, orientation),
				 xRadius, yRadius, orientAngle(225.0, orientation) - 48.0, 96.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);

	const BPoint points[3] = {
		mCenter + orientPoint(0, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -mSize, orientation)
	};
	StrokePolygon(points, 3, false, erase ? B_SOLID_LOW : B_SOLID_HIGH);
			
	const float rad = mSize * 0.5;
	StrokeArc(mCenter + orientPoint(0, -mSize, orientation),
				 rad, rad, orientAngle(315.0, orientation) - 48.0, 96.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeHHH(int orientation, bool erase)
{
	// The basic horizontal stroke.
	StrokeLine(mCenter + orientPoint(-mSize, -0.5 * mSize, orientation),
				  mCenter + orientPoint(mSize, -0.5 * mSize, orientation),
				  erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeHHV(int orientation, bool erase)
{
	// Horizontal stroke, horizontal to vertical.
	const BPoint points[3] = {
		mCenter + orientPoint(-mSize, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -0.5 * mSize, orientation),
		mCenter + orientPoint(0.5 * mSize, -mSize, orientation)
	};
	StrokePolygon(points, 3, false, erase ? B_SOLID_LOW : B_SOLID_HIGH);

	const float radius = mSize * 0.5;
	StrokeArc(mCenter + orientPoint(0, -mSize, orientation),
				 radius, radius, orientAngle(315.0, orientation) - 48.0, 96.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

void
KnotSectionCache::StrokeHVV(int orientation, bool erase)
{
	// "Horizontal" stroke, vertical to vertical.
	const float radius = mSize * 0.5;
	StrokeArc(mCenter + orientPoint(0, -mSize, orientation),
				 radius, radius, orientAngle(270.0, orientation) - 96.0, 192.0,
				 erase ? B_SOLID_LOW : B_SOLID_HIGH);
}

// Do antialiasing of mCanvas into mBuffer.
inline uint32
alpha(uint32 value)
{
	// This cryptic loop performs a "population count" (# of 1-bits).
	int numOnes = 0;
	while (value) {
		value -= (value & -value);
		++numOnes;
	}

	return (uint32)(255.0 * numOnes / 16.0);
}

void
KnotSectionCache::AntiAlias(void)
{
	const uint32 *pSrcBits = (const uint32 *)mCanvas.Bits();
	const int32 srcStride = mCanvas.BytesPerRow() / sizeof(uint32);
	
	uint32 *pDstBits = (uint32 *)mBuffer.Bits();
	const int32 dstStride = mBuffer.BytesPerRow() / sizeof(uint32);
	const uint32 pixel = (mColor.red << 16)
							 | (mColor.green << 8)
							 | mColor.blue;
	
	const BRect bounds = mBuffer.Bounds();
	const int32 width = bounds.IntegerWidth() + 1;
	const int32 height = bounds.IntegerHeight() + 1;

	for (int32 y = 0; y < height; ++y) {
		const uint32 *pSrc0 = pSrcBits;
		const uint32 *pSrc1 = pSrc0 + srcStride;
		const uint32 *pSrc2 = pSrc1 + srcStride;
		const uint32 *pSrc3 = pSrc2 + srcStride;
		uint32 *pDst = pDstBits;
		
		for (int32 x = 0; x < width;) {
			uint32 srcPix0 = B_BENDIAN_TO_HOST_INT32(*pSrc0++);
			uint32 srcPix1 = B_BENDIAN_TO_HOST_INT32(*pSrc1++);
			uint32 srcPix2 = B_BENDIAN_TO_HOST_INT32(*pSrc2++);
			uint32 srcPix3 = B_BENDIAN_TO_HOST_INT32(*pSrc3++);
			for (int32 pix = 0; pix < 8 && x < width; ++pix, ++x) {
				const uint32 n = (srcPix0 >> 28)
									|((srcPix1 >> 24) & 0x00f0)
									|((srcPix2 >> 20) & 0x0f00)
									|((srcPix3 >> 16) & 0xf000);				
				*pDst++ = B_HOST_TO_LENDIAN_INT32(pixel | (alpha(n) << 24));

				srcPix0 <<= 4;
				srcPix1 <<= 4;
				srcPix2 <<= 4;
				srcPix3 <<= 4;
			}
		}
		pSrcBits += srcStride * 4;
		pDstBits += dstStride;
	}
}
