//		Copyright © 1999–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Screen saver class.

#include "SamhainScreenSaver.h"

#include <Bitmap.h>
#include <DataIO.h>
#include <File.h>
#include <GraphicsDefs.h>
#include <image.h>
#include <Message.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <Resources.h>
#include <ScreenSaver.h>
#include <TranslationUtils.h>
#include <View.h>

#include "ClosedKnot.h"
#include "HorizontalKnot.h"
#include "RandomNumbers.h"
#include "Samhain.h"
#include "SamhainConfigView.h"
#include "SamhainPreferences.h"
#include "VerticalKnot.h"

namespace {
	// Constant data.
	const rgb_color kBlack = { 0, 0, 0, 255 };
	
	const bigtime_t kTick = 10000;					// 10msec per tick
	const bigtime_t kLogoInterval = 300000000;	// show logo every 5 minutes…
	const bigtime_t kLogoDuration = 5000000;		// …for 5 seconds
}

// Screen saver interface.
BScreenSaver *
instantiate_screen_saver(BMessage *pMsg, image_id id)
{
	return new SamhainScreenSaver(pMsg, id);
}

SamhainScreenSaver::SamhainScreenSaver(BMessage *pMsg, image_id id)
:	inherited(pMsg, id),
	mTimer(0),
	mLogoTimer(0),
	mpLogo(0),
	mLogoBounds(B_ORIGIN, B_ORIGIN),
	mLogoPosition(B_ORIGIN),
	mKnotsToDraw(0),
	mPreviewing(false),
	mShowingLogo(false)
{
	// Establish the user's preferences, or use defaults.
	if (!pMsg || mPreferences.LoadState(*pMsg) != B_OK) {
		mPreferences.SetDefaults();
	}
	
	// Try to be random.
	srand((long)system_time());

	// Try to load a logo bitmap from the add-on's resources.
	image_info info;
	if (get_image_info(id, &info) == B_OK) {
		BFile file(info.name, B_READ_ONLY);
		BResources resources(&file);
		size_t size;
		const void *pData =
			resources.LoadResource(B_RAW_TYPE, "BeOS logo data", &size);
		if (pData) {
			BMemoryIO stream(pData, size);
			mpLogo = BTranslationUtils::GetBitmap(&stream);
		}
		
		// ### For now, the alpha channel is stored in a separate bitmap.
		if (mpLogo && mpLogo->ColorSpace() == B_RGB32) {
			const void *pMask =
				resources.LoadResource(B_RAW_TYPE, "BeOS logo mask", &size);
			if (pMask) {
				BMemoryIO stream(pMask, size);
				BBitmap *pMaskBitmap = BTranslationUtils::GetBitmap(&stream);
				if (pMaskBitmap && pMaskBitmap->ColorSpace() == B_RGB32) {
					uint8 *pDataBits = (uint8 *)(mpLogo->Bits());
					uint8 *pMaskBits = (uint8 *)(pMaskBitmap->Bits());
					for (size_t n = mpLogo->BitsLength(); n >= sizeof(uint32); n -= sizeof(uint32)) {
						pDataBits[3] = pMaskBits[0];
						pDataBits += sizeof(uint32);
						pMaskBits += sizeof(uint32);
					}
				}
				delete pMaskBitmap;
			}
		}
	}
	
	if (mpLogo) {
		mLogoBounds = mpLogo->Bounds();
	}
}

SamhainScreenSaver::~SamhainScreenSaver(void)
{
	delete mpLogo;
	mpLogo = 0;
}

void
SamhainScreenSaver::Draw(BView *pView, int32 frame)
{
	// Wait out the time to the next scheduled event.
	if ((mTimer -= kTick) > 0 && frame > 0) {
		mLogoTimer -= kTick;
		return;
	}

	// Determine which set of preferences we'll use.
	const SamhainSaverPreferences &saverPreferences =
		mPreviewing ? mPreferences.PreviewPreferences()
						: mPreferences.SaverPreferences();

	const SamhainKnotPreferences &knotPreferences =
		mPreferences.KnotPreferences();
	
	// Draw something pretty.
	if (pView->LockLooper()) {
		const BRect bounds = pView->Bounds();
		
		// Clear screen if necessary.		
		if (--mKnotsToDraw <= 0 || frame == 0) {
			pView->SetHighColor(kBlack);
			pView->SetDrawingMode(B_OP_COPY);
			pView->FillRect(bounds);

			// Calculate new number of knots to be drawn.
			const double p = randomDouble();
			mKnotsToDraw = (int16)(			  p  * saverPreferences.MinNumKnots()
										  + (1.0 - p) * saverPreferences.MaxNumKnots());
		}

		// Determine knot type.
		KnotType type = kClosed;
		if (!mPreviewing) {
			const float p = randomFloat();
			if (p < knotPreferences.KnotProbability1()) {
				type = kHorizontal;
			} else if (p >= knotPreferences.KnotProbability2()) {
				type = kVertical;
			}
		}

		// ### TO DO:
		// ### - precalculate size (using preferences)

		// Draw the knot.			
		bool drawn;
		switch (type) {
		
		case kClosed:
			do {
				ClosedKnot knot(pView, knotPreferences);
				drawn = knot.Draw();
			} while (!(mPreviewing || drawn));
			break;

		case kHorizontal:
			do {
				HorizontalKnot knot(pView, knotPreferences);
				drawn = knot.Draw();
			} while (!drawn);
			break;

		case kVertical:
			do {
				VerticalKnot knot(pView, knotPreferences);
				drawn = knot.Draw();
			} while (!drawn);
			break;
		}

		// From time to time, give Be, Inc. some gratuitous screen time ;-)
		if ((mLogoTimer -= kTick) <= 0) {
			mShowingLogo = !mShowingLogo;
			mLogoTimer = mShowingLogo ? kLogoDuration : kLogoInterval;
			mLogoPosition =
				BPoint(randomFloat(bounds.Width() - mLogoBounds.Width()),
						 randomFloat(bounds.Height()- mLogoBounds.Height()));
		}
		
		if (!mPreviewing && mpLogo && mShowingLogo) {
			// Let's hear it for the BeOS!
			pView->SetDrawingMode(B_OP_ALPHA);
			pView->SetBlendingMode(B_PIXEL_ALPHA, B_ALPHA_OVERLAY);
			pView->DrawBitmap(mpLogo, mLogoPosition);
		}
		
		// OK, let go of the view.
		pView->UnlockLooper();
	}

	// Compute delay until next event.
	if (!mPreviewing && mpLogo && mShowingLogo) {
		// Override the usual behavior while the logo is shown.
		mTimer = mLogoTimer;

	} else {
		const double p = randomDouble();
		mTimer = (bigtime_t)(			p  * saverPreferences.MinDelay()
									+ (1.0 - p) * saverPreferences.MaxDelay());
	}
}

status_t
SamhainScreenSaver::SaveState(BMessage *pMsg) const
{
	// Save the user's current preferences.
	if (pMsg) {
		return mPreferences.SaveState(*pMsg);
	}
	return B_ERROR;
}

void
SamhainScreenSaver::StartConfig(BView *pConfigView)
{
	if (pConfigView->LockLooper()) {
		BRect frame = pConfigView->Bounds();
		frame.InsetBy(5.0, 0.0);
		pConfigView->AddChild(new SamhainConfigView(frame, mPreferences));
		pConfigView->UnlockLooper();
	}
}

status_t
SamhainScreenSaver::StartSaver(BView *, bool preview)
{
	SetTickSize(kTick);
	mTimer = 0;
	mLogoTimer = kLogoInterval;
	mKnotsToDraw = 0;
	mPreviewing = preview;
	mShowingLogo = false;
	
	return B_OK;
}
